# Copyright (C) 2022-2023 FormalGeo Development Team
# Author: Xiaokai Zhang
# Contact: formalgeo@gmail.com

"""Formal representation and solving for Euclidean plane geometry problems."""

__all__ = [
    "core", "problem", "solver", "tools"
]
